;****************************************************************
;*  d_timer.asm							*
;*			Countdown to a final date		*
;*			=========================		*
;*								*
;*  1998-10-04	Bodo Wenzel	Written DOWN2000, V1.0		*
;*  2009-02-17	Bodo Wenzel	V1.0, copied and edited for the	*
;*				new function			*
;*  2009-04-13	Bodo Wenzel	Edited for Alfred Arnold's	*
;*				assembler			*
;****************************************************************

	title		"Down Timer"

	relaxed	on	;literals in C notation

VERSN	equ	0x10

;================================================================
;	Definitions

	cpu	16c84
	include		"stddef18.inc"

;================================================================
;	Configuration

	segment	code
	org	0x2007
_CP_OFF		equ	0x3FFF
_PWRTE_ON	equ	0x3FFF
_WDT_OFF	equ	0x3FFB
_RC_OSC		equ	0x3FFF
	data	_CP_OFF&_PWRTE_ON&_WDT_OFF&_RC_OSC

;================================================================
;	Inputs and outputs

TRIS_A	equ	0b00010000	;port A:
;	equ	0..3		;O digit
HZ50	equ	4		;I 50Hz pulses

TRIS_B	equ	0b00000000	;Port B:
;	equ	0..7		;O segments

OPT	equ	0b11011111	;TMR0 as timer F_CYCLE/1

;================================================================
;	Variables
	segment	data

	org	0x0C

int_w:		res	1
int_status:	res	1
int_fsr:	res	1	;buffer for interrupt

flags:		res	1
fl_hz50		equ	0	;recognized level at HZ50
fl_key		equ	1	;key accepted at HZ50
fl_enter	equ	2	;enter mode active

debounce:	res	2

counter:	res	17	;digits

segments:	res	1	;buffer for segments switched on
digit:		res	1	;index into digits
point:		res	1	;index for decimal point in enter
				;mode

calc_count:	res	1
calc_bin:	res	2
calc_dec:	res	4	;scratchpad

;================================================================
;	Vectors
	segment	code

	org	0x0000
	goto	reset

	data	VERSN

	org	0x0004		;falls through...
;================================================================
;	Interrupt
interrupt:
	movwf	int_w
	swapf	STATUS,w
	bcf	STATUS,FLAG_RP0
	movwf	int_status
	movf	FSR,w
	movwf	int_fsr		;save all

	bcf	INTCON,RTIF	;reset timer
;----------------------------------------------------------------
	incf	digit
	movf	digit,w
	xorlw	10
	btfsc	STATUS,FLAG_Z
	clrf	digit		;count digit cyclically

	movf	digit,w
	addlw	counter
	movwf	FSR
	comf	INDF,w
	call	bin2ssg
	movwf	segments	;get next segments

	btfsc	flags,fl_enter
	goto	i_enter

	bsf	segments,2
	btfss	digit,3
	btfss	digit,0
	bcf	segments,2	;set decimal point (count down)
	goto	i_switch
i_enter:
	movf	digit,w
	andlw	0b11111110
	xorwf	point,w
	btfsc	STATUS,FLAG_Z
	bsf	segments,2	;set decimal point (enter mode)

i_switch:
	movlw	15
	movwf	PORTA
	movf	segments,w
	movwf	PORTB
	movf	digit,w
	movwf	PORTA		;switch on next digit
;----------------------------------------------------------------
	btfsc	PORTA,HZ50
	goto	i_hz50_h	;level "high"?

	incf	debounce+0
	btfss	debounce+0,2
	goto	i_quit		;"low" not yet debounced?

	decf	debounce+0
	bsf	flags,fl_hz50	;"low" recognized

	incfsz	debounce+1
	goto	i_quit		;key not yet accepted?

	bsf	flags,fl_key	;key accepted
	goto	i_quit

i_hz50_h:
	decf	debounce+0
	btfss	debounce+0,7
	goto	i_quit		;"high" not yet debounced?

	incf	debounce+0
	bcf	flags,fl_hz50	;"high" recognized
	clrf	debounce+1
	bcf	flags,fl_key	;key released
;----------------------------------------------------------------
i_quit:
	movf	int_fsr,w
	movwf	FSR
	swapf	int_status,w
	movwf	STATUS
	swapf	int_w
	swapf	int_w,w		;restore all
	retfie
;================================================================
;	Return a segment pattern for a decimal nibble
;	in:	W	nibble
;	out:	W	segment pattern
bin2ssg:
	addwf	PCL,f
		; fgbac.ed
	retlw	0b10111011	;0
	retlw	0b00101000	;1
	retlw	0b01110011	;2
	retlw	0b01111001	;3
	retlw	0b11101000	;4
	retlw	0b11011001	;5
	retlw	0b11011011	;6
	retlw	0b00111000	;7
	retlw	0b11111011	;8
	retlw	0b11111001	;9
;================================================================
;	Return the number of days for the month before
;	in:	W	month (0 = january, 11 = december)
;	out:	W	number of days
bin2month:
	addwf	PCL,f
	retlw	0	;jan / leap year
	retlw	31	;feb
	retlw	29	;mar
	retlw	31	;apr
	retlw	30	;may
	retlw	31	;jun
	retlw	30	;jul
	retlw	31	;aug
	retlw	31	;sep
	retlw	30	;oct
	retlw	31	;nov
	retlw	30	;dec
	retlw	0	;jan / normal year
	retlw	31	;feb
	retlw	28	;mar
	retlw	31	;apr
	retlw	30	;may
	retlw	31	;jun
	retlw	30	;jul
	retlw	31	;aug
	retlw	31	;sep
	retlw	30	;oct
	retlw	31	;nov
	retlw	30	;dec
;================================================================
;	Macros
tick	macro	label
tick_hl:
	btfsc	flags,fl_key
	goto	label		;key pressed?
	btfsc	flags,fl_hz50
	goto	tick_hl		;still "high"?
tick_lh:
	btfss	flags,fl_hz50
	goto	tick_lh		;still "low"?
	endm
;----------------------------------------------------------------
decr	macro	offs,reload,label
	incf	counter+offs,w
	btfsc	STATUS,FLAG_Z
	movlw	reload
	movwf	counter+offs
	btfss	STATUS,FLAG_Z
	goto	label		;decrement digit with limitation
	endm
;----------------------------------------------------------------
nokey	macro
loop:
	btfsc	flags,fl_key
	goto	loop		;wait for released key
	endm
;================================================================
;	Initialization
reset:
	bsf	STATUS,FLAG_RP0
	movlw	TRIS_A
	movwf	TRISA		;prepare port A
	movlw	TRIS_B
	movwf	TRISB		;prepare port B
	movlw	OPT
	movwf	OPTION		;set options
	bcf	STATUS,FLAG_RP0
;----------------------------------------------------------------
	movlw	bin2ssg >> 8
	movwf	PCLATH		;prepare table jumps

	clrf	flags
	clrf	debounce+0
	clrf	debounce+1
	clrf	digit		;initialize variables

	movlw	-6
	movwf	counter+1
	movlw	-1
	movwf	counter+0
	movwf	counter+2
	movwf	counter+3
	movwf	counter+4
	movwf	counter+5
	movwf	counter+6
	movwf	counter+7
	movwf	counter+8
	movwf	counter+9
	movwf	counter+10
	movwf	counter+16	;begin with 5 seconds left

	bsf	INTCON,RTIE
	bsf	INTCON,GIE
;================================================================
;	Main loop
main_loop:
	tick	enter_times	;wait for pulse or key

	decr	16,-5,main_loop	;prescale to 1/10s

	decr	0,-10,main_loop
	decr	1,-10,main_loop
	decr	2,-6,main_loop
	decr	3,-10,main_loop
	decr	4,-6,main_loop
	decr	5,-10,main_loop
	decr	6,-3,main_loop
	movlw	-4
	movwf	counter+5
	decr	7,-10,main_loop
	decr	8,-10,main_loop
	decr	9,-10,main_loop
	decr	10,-10,main_loop	;count down time left
	goto	main_loop
;----------------------------------------------------------------
;	First enter final date
enter_times:
	bsf	flags,fl_enter	;set enter mode

	movlw	-2
	movwf	counter+9
	movlw	-10
	movwf	counter+8
	movlw	-2
	movwf	counter+7
	movlw	-3
	movwf	counter+6
	movlw	-4
	movwf	counter+5
	movlw	-2
	movwf	counter+4
	movlw	-1
	movwf	counter+3
	movlw	-1
	movwf	counter+2
	movlw	-1
	movwf	counter+1
	movlw	-1
	movwf	counter+0	;set xx19-12-31 00:00
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
enter_year_then:
	movlw	8
	movwf	point		;set decimal point position

	nokey

	movlw	-50
	movwf	counter+16	;initialize prescaler
eyrt_loop:
	tick	enter_month_then	;wait for pulse or key

	decr	16,-50,eyrt_loop	;prescale to 1/1s

	decr	8,-10,eyrt_loop
	decr	9,-2,eyrt_loop
	goto	eyrt_loop	;count year cyclically 19..00
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
enter_month_then:
	movlw	6
	movwf	point		;set decimal point position

	nokey

	movlw	-50
	movwf	counter+16	;initialize prescaler
emtt_loop:
	tick	enter_day_then	;wait for pulse or key

	decr	16,-50,emtt_loop	;prescale to 1/1s

	incfsz	counter+7,w
	goto	emtt_check
	incf	counter+6
	incfsz	counter+6,w
	goto	emtt_loop
	movlw	-2
	movwf	counter+7
	movlw	-3
	movwf	counter+6
	goto	emtt_loop
emtt_check:
	decr	6,-10,emtt_loop
	incf	counter+7
	goto	emtt_loop	;count month cyclically 12..01
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
enter_day_then:
	movlw	4
	movwf	point		;set decimal point position

	nokey

	movlw	-50
	movwf	counter+16	;initialize prescaler
edyt_loop:
	tick	enter_datetime_now	;wait for pulse or key

	decr	16,-50,edyt_loop	;prescale to 1/1s

	incfsz	counter+5,w
	goto	edyt_check
	incf	counter+4
	incfsz	counter+4,w
	goto	edyt_loop
	movlw	-4
	movwf	counter+5
	movlw	-2		;no special handling for shorter
	movwf	counter+4	;months :-/
	goto	edyt_loop
edyt_check:
	decr	4,-10,edyt_loop
	incf	counter+5
	goto	edyt_loop	;count day cyclically 31..01
;----------------------------------------------------------------
;	Second enter current date and time
enter_datetime_now:
	movf	counter+9,w
	movwf	counter+15
	movf	counter+8,w
	movwf	counter+14
	movf	counter+7,w
	movwf	counter+13
	movf	counter+6,w
	movwf	counter+12
	movf	counter+5,w
	movwf	counter+11
	movf	counter+4,w
	movwf	counter+10	;save final date

	movlw	-2
	movwf	counter+9
	movlw	-10
	movwf	counter+8
	movlw	-2
	movwf	counter+7
	movlw	-3
	movwf	counter+6
	movlw	-4
	movwf	counter+5
	movlw	-2
	movwf	counter+4
	movlw	-3
	movwf	counter+3
	movlw	-4
	movwf	counter+2
	movlw	-6
	movwf	counter+1
	movlw	-10
	movwf	counter+0	;set xx19-12-31 23:59
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
enter_year_now:
	movlw	8
	movwf	point		;set decimal point position

	nokey

	movlw	-50
	movwf	counter+16	;initialize prescaler
eyrn_loop:
	tick	enter_month_now	;wait for pulse or key

	decr	16,-50,eyrn_loop	;prescale to 1/1s

	decr	8,-10,eyrn_loop
	decr	9,-2,eyrn_loop
	goto	eyrn_loop	;count year cyclically 19..00
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
enter_month_now:
	movlw	6
	movwf	point		;set decimal point position

	nokey

	movlw	-50
	movwf	counter+16	;initialize prescaler
emtn_loop:
	tick	enter_day_now	;wait for pulse or key

	decr	16,-50,emtn_loop	;prescale to 1/1s

	incfsz	counter+7,w
	goto	emtn_check
	incf	counter+6
	incfsz	counter+6,w
	goto	emtn_loop
	movlw	-2
	movwf	counter+7
	movlw	-3
	movwf	counter+6
	goto	emtn_loop
emtn_check:
	decr	6,-10,emtn_loop
	incf	counter+7
	goto	emtn_loop	;count month cyclically 12..01
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
enter_day_now:
	movlw	4
	movwf	point		;set decimal point position

	nokey

	movlw	-50
	movwf	counter+16	;initialize prescaler
edyn_loop:
	tick	enter_hour_now	;wait for pulse or key

	decr	16,-50,edyn_loop	;prescale to 1/1s

	incfsz	counter+5,w
	goto	edyn_check
	incf	counter+4
	incfsz	counter+4,w
	goto	edyn_loop
	movlw	-4
	movwf	counter+5
	movlw	-2
	movwf	counter+4
	goto	edyn_loop
edyn_check:
	decr	4,-10,edyn_loop
	incf	counter+5
	goto	edyn_loop	;count day cyclically 31..01
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
enter_hour_now:
	movlw	2
	movwf	point		;set decimal point position

	nokey

	movlw	-50
	movwf	counter+16	;initialize prescaler
ehrn_loop:
	tick	enter_minute_now	;wait for pulse or key

	decr	16,-50,ehrn_loop	;prescale to 1/1s

	decr	2,-10,ehrn_loop
	decr	3,-3,ehrn_loop
	movlw	-4
	movwf	counter+2
	goto	ehrn_loop	;count hour cyclically 23..00
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
enter_minute_now:
	movlw	0
	movwf	point		;set decimal point position

	nokey

	movlw	-50
	movwf	counter+16	;initialize prescaler
emin_loop:
	tick	enter_ready	;wait for pulse or key

	decr	16,-50,emin_loop	;prescale to 1/1s

	decr	0,-10,emin_loop
	decr	1,-6,emin_loop
	goto	emin_loop	;count minute cyclically 59..00
;----------------------------------------------------------------
;	Finally calculate days and time left
enter_ready:
	movlw	4
	call	input2days	;calculate # of days for "now"
	movf	calc_bin+0,w
	movwf	calc_dec+0
	movf	calc_bin+1,w
	movwf	calc_dec+1

	incf	calc_dec+1
	btfsc	STATUS,FLAG_Z
	incf	calc_dec+0	;the day has already begun...

	movlw	10
	call	input2days	;calculate # of days for "then"

	movf	calc_dec+1,w
	subwf	calc_bin+1,f
	btfss	STATUS,FLAG_C
	decf	calc_bin+0
	movf	calc_dec+0,w
	subwf	calc_bin+0,f	;subtract now" from "then"

	btfsc	STATUS,FLAG_C
	goto	er_nowrap
	movlw	(20*365+5)&0xFF
	addwf	calc_bin+1,f
	btfsc	STATUS,FLAG_C
	incf	calc_bin+0
	movlw	(20*365+5)>>8
	addwf	calc_bin+0,f	;wrap around
er_nowrap:

	call	bin2dec
	movf	calc_dec+1,w
	movwf	counter+9
	movf	calc_dec+2,w
	movwf	counter+8
	movf	calc_dec+3,w
	movwf	counter+7	;put left days into counter

	clrf	calc_bin+0
	comf	counter+2,w
	btfss	counter+3,0
	addlw	10
	btfss	counter+3,1
	addlw	20
	sublw	23
	movwf	calc_bin+1	;calculate left hours

	call	bin2dec
	movf	calc_dec+2,w
	movwf	counter+6
	movf	calc_dec+3,w
	movwf	counter+5	;put left hours into counter

	clrf	calc_bin+0
	comf	counter+0,w
	btfss	counter+1,0
	addlw	10
	btfss	counter+1,1
	addlw	20
	btfss	counter+1,2
	addlw	40
	sublw	59
	movwf	calc_bin+1	;calculate left minutes

	call	bin2dec
	movf	calc_dec+2,w
	movwf	counter+4
	movf	calc_dec+3,w
	movwf	counter+3	;put left minutes into counter

	movlw	-6
	movwf	counter+2
	movlw	-10
	movwf	counter+1
	movlw	-10
	movwf	counter+0
	movlw	-1
	movwf	counter+16	;fill up seconds and prescaler

	nokey

	bcf	flags,fl_enter	;set counter mode
	goto	main_loop
;================================================================
;	Calculate days from date input
;	in:	W		offset into "counter"
;	out:	calc_bin	number of days
input2days:
	addlw	counter
	movwf	FSR		;set pointer

	clrf	calc_bin+0
	comf	INDF,w
	incf	FSR
	btfss	INDF,0
	addlw	10
	btfss	INDF,1
	addlw	20
	incf	FSR
	movwf	calc_bin+1	;fetch day

	comf	INDF,w
	incf	FSR
	btfss	INDF,0
	addlw	10
	incf	FSR
	movwf	counter+16	;fetch month

	comf	INDF,w
	incf	FSR
	btfss	INDF,0
	addlw	10
	movwf	calc_count	;fetch year

	movlw	12
	btfss	calc_count,0
	btfsc	calc_count,1
	addwf	counter+16,f	;add offset for normal year

e2d_month_loop:
	decf	counter+16
	movf	counter+16,w
	bcf	INTCON,RTIE
	call	bin2month
	bsf	INTCON,RTIE
	addwf	calc_bin+1,f
	btfsc	STATUS,FLAG_C
	incf	calc_bin+0
	iorlw	0
	btfss	STATUS,FLAG_Z
	goto	e2d_month_loop	;add days of the full months

	incf	calc_count	;since there could be year "0"
e2d_year_loop:
	movlw	366&0xFF
	btfss	calc_count,0
	btfss	calc_count,1
	movlw	365&0xFF
	addwf	calc_bin+1,f
	btfsc	STATUS,FLAG_C
	incf	calc_bin+0
	movlw	365>>8		;identical to 366>>8
	addwf	calc_bin+0,f
	decfsz	calc_count
	goto	e2d_year_loop	;add days for the years

	return
;================================================================
;	Calculate BCD nibbles from binary (only 4 digits result!)
;	in:	calc_bin	binary number
;	out:	calc_dec	BCD nibbles
bin2dec:
	clrf	calc_dec+0
	clrf	calc_dec+1

	movlw	16
	movwf	calc_count
b2d_loop:
	movlw	0x03
	addwf	calc_dec+0,f
	btfss	calc_dec+0,3
	subwf	calc_dec+0,f
	addwf	calc_dec+1,f
	btfss	calc_dec+1,3
	subwf	calc_dec+1,f
	movlw	0x30
	addwf	calc_dec+0,f
	btfss	calc_dec+0,7
	subwf	calc_dec+0,f
	addwf	calc_dec+1,f
	btfss	calc_dec+1,7
	subwf	calc_dec+1,f	;BCD correction

	rlf	calc_bin+1
	rlf	calc_bin+0
	rlf	calc_dec+1
	rlf	calc_dec+0	;shift 1 bit

	decfsz	calc_count
	goto	b2d_loop	;not yet all bits done?

	comf	calc_dec+1,w
	iorlw	0xF0
	movwf	calc_dec+3
	swapf	calc_dec+1,w
	xorlw	0xFF
	iorlw	0xF0
	movwf	calc_dec+2
	comf	calc_dec+0,w
	iorlw	0xF0
	movwf	calc_dec+1
	swapf	calc_dec+0,w
	xorlw	0xFF
	iorlw	0xF0
	movwf	calc_dec+0	;separate BCD nibbles
	return
;================================================================
	end
